#include	<Canbus.h>
#include	<defaults.h>
#include	<global.h>
#include	<mcp2515.h>
#include	<mcp2515_defs.h>
#include	<Wire.h>
#include	<TimerOne.h>			// TimerOne library

#define GIMBAL_INIT			-8		// unit: deg

#define CANSPEED_125_16M	7		// CAN speed at 125 kbps@16M osc  (CNF1 BRP, 16/Fosc)
#define CANSPEED_250_16M	3		// CAN speed at 250 kbps@16M osc  (CNF1 BRP, 8/Fosc)
#define CANSPEED_500_16M	1		// CAN speed at 500 kbps@16M osc  (CNF1 BRP, 4/Fosc)
//
#define CANSPEED_125_8M		3		// CAN speed at 125 kbps@8M osc   (CNF1 BRP, 8/Fosc)
#define CANSPEED_250_8M		1		// CAN speed at 250 kbps@8M osc   (CNF1 BRP, 4/Fosc)
#define CANSPEED_500_8M		0		// CAN speed at 500 kbps@8M osc   (CNF1 BRP, 2/Fosc)

#define GYRO_OFFSET_MA		500

#define MPU6050_ADR			0x68	// I2C address of the MPU-6050 & MPU-9250
#define BMI160_ADR			0x69
#define BNO055_ADR			0x29

//#define USE_MPU6050		// or MPU9250
//#define USE_BMI160
#define USE_BNO055

int	servo_pin	= 9;

///////////////////////////////////
// I2C
///////////////////////////////////
void i2c_write(int dev_adr, int reg_adr, int val)
{
	Wire.beginTransmission(dev_adr);
	Wire.write(reg_adr);
	Wire.write(val);
	Wire.endTransmission(true);
}

int i2c_read_1byte(int dev_adr, int reg_adr)
{
	int	ret;
	Wire.beginTransmission(dev_adr);
	Wire.write(reg_adr);
	Wire.endTransmission(false);
	Wire.requestFrom(dev_adr, 1, true);
	ret	= Wire.read() & 0xFF;
	return ret;
}


int i2c_read_2byte(int dev_adr, int reg_adr)
{
	int	ret;
	Wire.beginTransmission(dev_adr);
	Wire.write(reg_adr);
	Wire.endTransmission(false);
	Wire.requestFrom(dev_adr, 2, true);
	ret	= Wire.read() & 0xFF;
	ret	<<= 8;
	ret	+= (Wire.read() & 0xFF);
	return ret;
}


int i2c_read_2byte_LH(int dev_adr, int reg_adr)
{
	int	ret;
	Wire.beginTransmission(dev_adr);
	Wire.write(reg_adr);
	Wire.endTransmission(false);
	delay(2);						// for BNO055
	Wire.requestFrom(dev_adr, 2, true);
	ret	= Wire.read() & 0xFF;
	ret	|= ((Wire.read() & 0xFF) << 8);
	return ret;
}


///////////////////////////////////
// timer interrupt vector
///////////////////////////////////
volatile int	pwm_cnt;
volatile int	pwm_h_period;

volatile int	sns_cnt;
volatile int	sns_flag;


void timer1_func()		// 100us
{
	pwm_cnt++;
	if (pwm_cnt >= 200) {	// 20ms
		digitalWrite(servo_pin, HIGH);
		pwm_cnt	= 0;
	}
	else if (pwm_cnt >= pwm_h_period) {
		digitalWrite(servo_pin, LOW);
	}

	sns_cnt++;
	if (sns_cnt >= 100) { // 10ms
		sns_flag  = 1;
		sns_cnt   = 0;
	}
}


///////////////////////////////////
// main
///////////////////////////////////
long	attitude;
long	gyro_offset;
long	gyro_z;
#ifdef USE_BNO055
long	attitude_offset;
#endif

int		cansend_cnt;


int gimbal_agl_to_pwm(int cmd_deg)
{
	int ret;
	// -45deg to 45deg corresponds to 0.95ms(9) to 1.95ms(19)
	ret = 14 + ((cmd_deg * 5) / 40);
	return ret;
}


void setup()
{
	pinMode(servo_pin, OUTPUT);

	// set timer
	pwm_cnt			= 0;
	pwm_h_period	= gimbal_agl_to_pwm(GIMBAL_INIT);
	sns_cnt			= 0;
	sns_flag		= 0;

	Timer1.initialize(100);					// unit: us
	Timer1.attachInterrupt(timer1_func);

	// config CAN
	Canbus.init(CANSPEED_250_8M);			// ext CAN brd
//	Canbus.init(CANSPEED_250_16M);			// CAN shield

  // init gyro
  Wire.begin();

#ifdef USE_MPU6050		// (or MPU9250)
	i2c_write(MPU6050_ADR, 0x6B, 0x80);		// reset device
	delay(150);
	i2c_write(MPU6050_ADR, 0x6A, 0x05);		// reset sig cond
	delay(150);
	i2c_write(MPU6050_ADR, 0x6B, 0x00);		// exit sleep mode
	delay(150);
	i2c_write(MPU6050_ADR, 0x1A, 0x01);		// 188Hz filter
	i2c_write(MPU6050_ADR, 0x1B, 0x00);		// gyro 250dps mode
	delay(150);
#endif
#ifdef USE_BMI160
	i2c_write(BMI160_ADR, 0x7E, 0xB6);		// reset device
	delay(100);
	i2c_write(BMI160_ADR, 0x7E, 0x15);		// gyro to normal mode
	delay(100);
	i2c_write(BMI160_ADR, 0x42, 0x29);		// gyro ODR 200Hz
	i2c_write(BMI160_ADR, 0x43, 0x03);		// gyro 250dps mode
#endif
#ifdef USE_BNO055
	i2c_write(BNO055_ADR, 0x07, 0x00);		// page 0
	i2c_write(BNO055_ADR, 0x3F, 0xC0);		// reset
	delay(100);
	i2c_write(BNO055_ADR, 0x3E, 0x00);		// normal mode
	i2c_write(BNO055_ADR, 0x3D, 0x0C);		// NDOF
	i2c_write(BNO055_ADR, 0x3B, 0x80);		// output format
	i2c_write(BNO055_ADR, 0x07, 0x00);		// page 0
#endif

	// init global var
	attitude	= 0;
	gyro_offset	= 0;
	cansend_cnt	= 0;
}


void loop()
{
	tCAN	message;
	int		cmd_deg;

	// receive CAN command
	if (mcp2515_check_message()) {
		if (mcp2515_get_message(&message)) {
			// gimbal command
			if (message.id == 0x100) {
				cmd_deg	= message.data[0] & 0xFF;

				// to signed
				if (cmd_deg >= 0x80) {
					cmd_deg	-= 0x100; // -127 - 127
				}

				// limit
				if (cmd_deg > 45) {
					cmd_deg	= 45;
				}
				if (cmd_deg < -45) {
					cmd_deg	= -45;
				}

				// set servo pwm
				pwm_h_period	= gimbal_agl_to_pwm(cmd_deg);
			}
			// zero reset command
			else if (message.id == 0x101) {
				int		i;
				long	sum;

				pwm_h_period	= gimbal_agl_to_pwm(GIMBAL_INIT);	// stop movement
				delay(1000);

				gyro_offset	= 0;
				sum			= 0;
				for (i = 0; i < GYRO_OFFSET_MA; i++) {
					sns_flag	= 0;
					while (sns_flag == 0)
						;
#ifdef USE_MPU6050
					gyro_z	= (long)i2c_read_2byte(MPU6050_ADR, 0x47);
					if (gyro_z > 0x8000) {
						gyro_z	-= 0x10000;
					}
#endif
#ifdef USE_BMI160
					gyro_z	= (long)i2c_read_2byte_LH(BMI160_ADR, 0x10);
					if (gyro_z > 0x8000) {
						gyro_z	-= 0x10000;
					}
#endif
#ifdef USE_BNO055
					gyro_z	= (long)i2c_read_2byte_LH(BNO055_ADR, 0x18);
					if (gyro_z > 0x8000) {
						gyro_z	-= 0x10000;
					}
					gyro_z	*= 10;		// note: sign, CCW
					gyro_z	/= 16;		// now 0.1deg/LSB
					gyro_z	*= 13;		// 0.007deg/LSB (corresponds to 250dps range)
#endif
					sum += gyro_z;
				}
				sum			/= GYRO_OFFSET_MA;
				gyro_offset	= sum;

#ifdef USE_BNO055
				attitude	= (long)i2c_read_2byte_LH(BNO055_ADR, 0x1A);	// unit: 1/16deg
				if (attitude > 0x8000) {
					attitude	-= 0x10000;
				}
				attitude		*= -10;	// note: sign, CCW
				attitude		/= 16;	// now 0.1deg/LSB
				attitude_offset	= attitude;
#endif
				attitude	= 0;
			}
		}
	}

	// check attitude (10ms interval)
	if (sns_flag == 1) {
		sns_flag	= 0;

#ifdef USE_MPU6050
		gyro_z	= (long)i2c_read_2byte(MPU6050_ADR, 0x47);
		if (gyro_z > 0x8000) {
			gyro_z	-= 0x10000;
		}
#endif
#ifdef USE_BMI160
		gyro_z	= (long)i2c_read_2byte_LH(BMI160_ADR, 0x10);
		if (gyro_z > 0x8000) {
			gyro_z	-= 0x10000;
		}
#endif
#ifdef USE_BNO055
		gyro_z	= (long)i2c_read_2byte_LH(BNO055_ADR, 0x18);
		if (gyro_z > 0x8000) {
			gyro_z	-= 0x10000;
		}
		gyro_z	*= 10;			// note: sign, CCW
		gyro_z	/= 16;			// now 0.1deg/LSB
		gyro_z	*= 13;			// 0.007deg/LSB (corresponds to 250dps range)
#endif

#ifndef USE_BNO055
		attitude	+= ((gyro_z - gyro_offset) * 1000) / 100;		// x1000 value; /100 corresponds to 10ms interval (*0.01)
#else	// BNO055
		attitude	= (long)i2c_read_2byte_LH(BNO055_ADR, 0x1A);	// unit: 1/16deg
		if (attitude > 0x8000) {
			attitude	-= 0x10000;
		}
		attitude	*= -10;		// note: sign, CCW
		attitude	/= 16;		// now 0.1deg/LSB
		// adjust offset
		attitude	-= attitude_offset;
		// adjust range
		while (attitude > 1800) {
			attitude	-= 3600;
		}
		while (attitude <= -1800) {
			attitude  += 3600;
		}
		// adjust scale
		attitude	*= 13;		// 0.007deg/LSB
		// extend
		attitude	*= 1000;	// x1000 value
#endif

		// send status (40ms interval)
		cansend_cnt++;
		if (cansend_cnt >= 4) {
			cansend_cnt	= 0;

			message.id				= 0x102;
			message.header.rtr		= 0;
			message.header.length	= 8;
			message.data[0]			= (gyro_z >> 8) & 0xFF;
			message.data[1]			= gyro_z & 0xFF;
			message.data[2]			= (gyro_offset >> 8) & 0xFF;
			message.data[3]			= gyro_offset & 0xFF;
			message.data[4]			= (attitude >> 24) & 0xFF;
			message.data[5]			= (attitude >> 16) & 0xFF;
			message.data[6]			= (attitude >> 8) & 0xFF;
			message.data[7]			= attitude & 0xFF;
			mcp2515_bit_modify(CANCTRL, (1<<REQOP2)|(1<<REQOP1)|(1<<REQOP0), 0);
			mcp2515_send_message(&message);
		}
	}
}
// end of file
