#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#define	OUT_SUFFIX	".csv"

#define	SERVO_OFFSET	-8

#define	GYRO_LSB	0.00763358778625954198473282442748f		// 250dps mode


int main(int argc, char *argv[])
{
	FILE	*fpin, *fpout;
	int		fnum;
	char	fname[1024];

	int		cmd;
	double	dps;

	if (argc < 2) {
		fprintf(stderr, "usage: read_candump file1 file2 file3 ...\n");
		exit(1);
	}

	// iterate for all input files
	for (fnum = 1; fnum < argc; fnum++) {
		char	linebuf[1024];

		// open input
		if ((fpin = fopen(argv[fnum], "rt")) == (FILE *)NULL) {
			fprintf(stderr, "SKIP: cannot open %s\n", argv[fnum]);
			continue;
		}

		// create output
		sprintf(fname, "%s%s", argv[fnum], OUT_SUFFIX);
		if ((fpout = fopen(fname, "wt")) == (FILE *)NULL) {
			fprintf(stderr, "SKIP: cannot create %s\n", fname);
			fclose(fpin);
			continue;
		}

		// read candump
		cmd	= SERVO_OFFSET;
		dps	= 0.0f;

		while (fgets(linebuf, 1024 - 10, fpin) != NULL) {
			double	tm;
			int		id, v0, v1, v2, v3, v4, v5, v6, v7, len;
			int		gz, cm;

			sscanf(linebuf, " (%lf)  can0  %03X   [%d]  %02X %02X %02X %02X %02X %02X %02X %02X\n",
						&tm, &id, &len, &v0, &v1, &v2, &v3, &v4, &v5, &v6, &v7
					);

			switch (id) {
			case 0x101:
				cm	= v0 & 0xFF;
				if (cm >= 0x80)
					cm	-= 0x100;
				cmd	= cm;
				break;

			case 0x102:
				gz	= ((v1 & 0xFF) << 8) | (v2 & 0xFF);
				if (gz >= 0x8000)
					gz	-= 0x10000;
				dps	= (double)gz * GYRO_LSB;

				fprintf(fpout, "%d,%f\n", cmd, dps);
				break;

			default:
				break;
			}
		}

		fclose(fpin);
		fclose(fpout);
	}

	return 0;
}
// EOF
